from contextlib import contextmanager
import requests
import subprocess
import json

@contextmanager
def safe_popen(*args, **kwargs):
    """Context manager that works exactly like subprocess.Popen, except that
    the process is killed at the end of the with statement.

        with safe_popen("python webapp.py 8080"):
            time.sleep(0.5) # wait for the app to start
            assert urlopen("http://localhost:8080/").code == 200
        # The webserver will be automatically killed here

    """
    p = subprocess.Popen(*args, **kwargs)
    try:
        yield p
    finally:
        p.kill()

class API:
    def __init__(self, base_url):
        self.base_url = base_url

    def request(self, method, path, data=None):
        url = self.base_url + path
        if isinstance(data, dict):
            data = json.dumps(data).encode('utf-8')
        return requests.request(method, url, data=data)

    def get(self, path):
        return self.request('GET', path)

    def post(self, path, data):
        return self.request('POST', path, data=data)

    def delete(self, path):
        return self.request('DELETE', path)

    def put(self, path, data):
        return self.request('PUT', path, data=data)

    def patch(self, path, data):
        return self.request('PATCH', path, data=data)

class Something:
    """Something is a special object that returns True
    when compared to any object except None.
    """
    def __eq__(self, other):
        return other is not None

    def __repr__(self):
        return "<Something>"




import logging
import json
import os
import functools
import inspect
import csv
from io import BytesIO as StringIO

logger = logging.getLogger("base")

def to_kwargs(f, *args, **kwargs):
    """Takes arguments given to a function and converts all of them into
    keyword arguments by looking at the function signature.
    
    >>> def f(a, b=2): pass
    ...
    >>> to_kwargs(f, 1)
    {'a': 1, 'b': 2}
    >>> to_kwargs(f, 1, 3)
    {'a': 1, 'b': 3}
    >>> to_kwargs(f, b=3, a=1)
    {'a': 1, 'b': 3}
    """

    s = inspect.getargspec(f)
    defaults = s.defaults or []
    default_args = s.args[-len(defaults):]

    kw = {}
    kw.update(zip(default_args, defaults))
    kw.update(kwargs)
    kw.update(zip(s.args, args))
    return kw

def to_args(f, *args, **kwargs):
    """Takes arguments given to a function and converts all of them into
    positional arguments by looking at the function signature.

    >>> def f(a, b=2): pass
    ...
    >>> to_args(f, 1)
    [1, 2]
    >>> to_args(f, 1, 3)
    [1, 3]
    >>> to_args(f, b=3, a=1)
    [1, 3]
    """
    kwargs = to_kwargs(f, *args, **kwargs)
    s = inspect.getargspec(f)
    return [kwargs[a] for a in s.args]

class DiskCache(object):
    """Disk-based cache.

    Provides a memoize decorator.
    """
    def __init__(self, root):
        self.root = root

    def memoize(self, path):
        """Memoize decorator that uses disk as cache.

        If the path ends with .json, the return value is encoded into JSON
        before saving and decoded on read.
        
        String formatting opeator can be used in the path, the actual path
        is constructed by formatting the specified path using the argument
        values.
        
        Usage:

            cache = DiskCache("cache/")
        
            @cache.memoize("data/c_{number:02d}")
            def get_consititency(self, number):
                ...

            @cache.memoize("data/c_{1:02d}")
            def get_consititency(self, number):
                ...

        """
        path = os.path.join(self.root, path)
        def decorator(f):
            @functools.wraps(f)
            def g(*a, **kw):
                kwargs = to_kwargs(f, *a, **kw)
                args = to_args(f, *a, **kw)
                filepath = path.format(*args, **kwargs)            
                disk = Disk()
                content = disk.read(filepath)
                if content:
                    return content
                else:
                    content = f(*a, **kw)
                    content = disk.write(filepath, content)
                    return disk.read(filepath)
            return g
        return decorator

def safestr(x):
    if isinstance(x, (list, set)):
        return [safestr(a) for a in x]
    if isinstance(x, unicode):
        return x.encode('utf-8')
    else:
        return x

def safeunicode(x):
    if isinstance(x, (list, set)):
        return [safeunicode(a) for a in x]
    if isinstance(x, str):
        return x.decode('utf-8')
    else:
        return x

class Disk:
    """Simple wrapper to read and write files in various formats.
    
    This takes care of coverting the data to and from the required format. The default format is text.

    Other supported formats are:
        * json
    """
    def write(self, path, content):
        print("write({0!r}, {1!r})".format(path, content))
        if path.endswith(".json"):
            if inspect.isgenerator(content):
                content = list(content)
            content = json.dumps(content, indent=4)
        elif path.endswith(".csv"):
            f = StringIO()
            w = csv.writer(f)
            data = [safeunicode(row) for row in content]
            w.writerows(data)
            content = f.getvalue()
        elif path.endswith(".tsv"):
            f = StringIO()
            w = csv.writer(f, delimiter="\t")
            data = [safestr(row) for row in content]
            w.writerows(data)
            content = f.getvalue()
            
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        
        logger.info("saving %s", path)
        with open(path, 'w') as f:
            f.write(content)
    
    def read(self, path):
        if os.path.exists(path):
            logger.info("reading %s", path)
            f = open(path)
            if path.endswith(".json"):
                return json.load(f)
            elif path.endswith(".csv"):
                reader = csv.reader(f)
                return list(reader)
            elif path.endswith(".tsv"):
                reader = csv.reader(f, delimiter="\t")
                return list(reader)
            else:
                return f.read()
            
def setup_logger():            
    FORMAT = "%(asctime)s [%(name)s] [%(levelname)s] %(message)s"
    logging.basicConfig(format=FORMAT, level=logging.INFO)



"""
Tool to render a Jinja template using values in the environment.
"""
from __future__ import print_function
import argparse
import jinja2
import os

def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument("template", help="template to render")
    return p.parse_args()

def main():
    args = parse_args()
    t = jinja2.Template(open(args.template).read(), undefined=jinja2.StrictUndefined)
    print(t.render(**os.environ))

if __name__ == '__main__':
    main()



"""Simple browser like interface for nagivating a web page and extracting
data. 

Uses requests and BeautifulSoup.
"""
from bs4 import BeautifulSoup
import requests
import logging

logger = logging.getLogger(__name__)

class Browser:
    def __init__(self, start_url):
        self._session = requests.Session()
        self._session.headers['referer'] = start_url
        self.reset()
        self.start_url = start_url

    def reset(self):
        self._response = None
        self._soup = None        

    def get(self, url, **kw):
        logger.info("GET %s", url)
        self._response = self._session.get(url, **kw)
        return self._response

    def post(self, url, data, **kw):
        logger.info("POST %s", url)
        self._response = self._session.post(url, data, **kw)
        return self._response

    def get_soup(self):
        """Returns beautiful soup of the current document."""
        if self._soup is None:
            if self._response is None:
                self.get(self.start_url)
            self._soup = BeautifulSoup(self._response.text, "lxml", from_encoding='utf8')
        return self._soup

    def get_form(self, **kw):
        form = self.get_soup().find("form", **kw)
        return form and Form(self, form)

    def read_formdata(self, inputs):
        soup = self.get_soup()
        params = {}

        for s in soup.findAll("select"):
            params[s['name']] = self.find_select_value(s)

        #params['GoogleMapForASPNet1$hidEventName'] = ''
        #params['GoogleMapForASPNet1$hidEventValue'] = ''
        params['__EVENTARGUMENT'] = ''
        params['__LASTFOCUS'] = ''
        ev = soup.find("input", {"name": "__EVENTVALIDATION"})
        if ev:
            params['__EVENTVALIDATION']  = ev['value']
        params['__VIEWSTATE'] = soup.find("input", {"name": "__VIEWSTATE"})['value']
        return params

    def get_select_options(self, name, ignore=None):
        select = self.get_soup().find("select", {"name": name})
        options = select.find_all("option")
        d = dict((o['value'], o.get_text().strip()) for o in options)

        if ignore:
            for k in ignore:
                d.pop(k, None)
        return d

class Form:
    def __init__(self, browser, form_element):
        self._browser = browser
        self._form = form_element
        self._data = None
        self.action = self._browser._response.url

    def __getitem__(self, name):
        return self.data[name]

    def __setitem__(self, name, value):
        self.data[name] = value

    def __delitem__(self, name):
        del self.data[name]

    @property
    def data(self):
        if self._data is None:
            self._data = self.get_data()
        return self._data            

    def get_data(self):
        soup = self._form
        params = {}

        for s in soup.findAll("select"):
            params[s['name']] = self.find_select_value(s) or ""

        for s in soup.findAll("input"):
            params[s['name']] = s['value'] or ""
        return params

    def find_select_value(self, select):
        if isinstance(select, basestring):
            select = self._form.find("select", {"name": select})
        option = select.find("option", {"selected": "selected"})
        return option and option.get('value')

    def submit(self):
        self._browser.post(self.action, self.data, headers={'referer': self.action})

    def __repr__(self):
        return "<Form {}>".format(self.data)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

requirements = [
    'Jinja2'
]

setup(
    name='niftyhacks',
    version='0.2.0',
    description='assorted hacks by @anandology',
    author='Anand Chitipothu',
    author_email='anandology@gmail.com',
    url='https://github.com/anandology/niftyhacks',
    packages=[
        'niftyhacks',
    ],
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    entry_points='''
        [console_scripts]
        jinja-envy=niftyhacks.jinja_envy:main
    '''
)



"""DB utilities to work with web.py databases.

This module provides utilities to inspect the database schema to find the
available tables, coulmns and constraints.

This is tested only on postgres db.
"""
from collections import defaultdict

class Schema:
    """Database Schema

    Provides access tables in the database.
    """
    def __init__(self, db):
        self.db = db

    def get_tables(self, table_schema='public', **filters):
        """Returns all tables in the database.

        :param table_schema: specified the table_schema to list tables in
        :return: list of table objects
        """
        rows = self.db.where("information_schema.tables", table_schema=table_schema, **filters)
        return [Table(self.db, row) for row in rows]

    def get_table(self, table_name, table_schema="public"):
        """Returns the table with specified table name.

        :param table_name: name of the table
        :param table_schema: name of the table schema
        :return: the table object
        """
        tables = self.get_tables(table_name=table_name, table_schema=table_schema)
        return tables and tables[0] or None

    def has_table(self, table_name, table_schema='public'):
        return self.get_table(table_name=table_name, table_schema=table_schema) is not None

    def get_enum_types(self):
        return EnumType.find_all(self.db)

    def get_enum_type(self, type_name):
        return EnumType.find(self.db, type_name)

    def has_enum_type(self, type_name):
        return self.get_enum_type(type_name) is not None

class Table:
    """Table in a database.

    Important Attributes:
        - table_schema
        - table_name
        - table_type (VIEW, BASE TABLE)
    """
    def __init__(self, db, table_data):
        self.db = db
        self.__dict__.update(table_data)

    def get_columns(self, **filters):
        rows = self.db.where("information_schema.columns",
                table_schema=self.table_schema,
                table_name=self.table_name,
                **filters)
        return [Column(self, row) for row in rows]

    def get_column(self, column_name):
        columns = self.get_columns(column_name=column_name)
        return columns and columns[0] or None

    def has_column(self, column_name):
        return self.get_column(column_name) is not None

class Column:
    """A column in a database table.

    Important attributes:

        - column_name
        - data_type
        - column_default
        - is_nullable
    """
    def __init__(self, table, column_data):
        self.table = table
        self.__dict__.update(column_data)

class EnumType:
    """EnumType in postgres database.
    """
    def __init__(self, name, values):
        self.name = name
        self.values = values

    def __eq__(self, other):
        return isinstance(other, EnumType) \
            and self.name == other.name \
            and self.values == other.values

    def __repr__(self):
        return "<TYPE {} ENUM {}>".format(self.name, self.values)

    @classmethod
    def find_all(cls, db):
        q = """
        SELECT
            n.nspname as enum_schema,
            t.typname as enum_name,
            e.enumlabel as enum_value,
            e.enumsortorder as sort_order
        FROM pg_type t
        JOIN pg_enum e ON t.oid = e.enumtypid
        JOIN pg_catalog.pg_namespace n ON n.oid = t.typnamespace
        WHERE n.nspname = 'public'
        ORDER BY t.typname, e.enumsortorder
        """
        types = defaultdict(list)
        for row in db.query(q):
            types[row.enum_name].append(row.enum_value)
        return [EnumType(name, values) for name, values in types.items()]

    @classmethod
    def find(cls, db, name):
        for enum in cls.find_all(db):
            if enum.name == name:
                return enum
